#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#
#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, Any, List, Dict, Tuple, Protocol

import threading
import queue
import copy


class Callback(Protocol):
    def __call__(self, name: str, item: Any) -> bool:
        ...  # pragma: no cover


class PluggableNotifier(threading.Thread):
    """
    ``PluggableNotifier`` implements an observer pattern.
    Subjects can register tuples consisting of executable filter expressions
    and callbacks to be notified about events.

    There is a method to enqueue new items and a thread to
    apply filters on items. Communication between
    method and thread is done via a queue.

    ``PluggableNotifier`` extends :class:`threading.Thread`.
    """

    def __init__(
        self,
        verbosity: int = 0,
        raise_exceptions: bool = False,
        filter_exception_handler: Optional[Callback] = None,
        callback_exception_handler: Optional[Callback] = None,
    ) -> None:
        """
        Args:
            verbosity: if set to 1 debug messages are printed to stdout
            raise_exceptions: if set to ``True`` exceptions are raised
                if a filter expression or a callback raises an exception.
                This may be used for debugging. In case of an exception
                :func:`~htf.communication.PluggableNotifier.stop` is called so
                no further items can be put in.
            filter_exception_handler: a callback that is
                called to handle a raised exception in a filter expression.
                If the ``filter_exception_handler`` returns ``True`` the
                exception is not raised afterwards. Else the exception handler
                raises the catched exception and stopps the thread.
                The supplied handler musst accept ``name`` and ``item``
                arguments. ``name`` is the callback name supplied by
                :meth:`~htf.communication.PluggableNotifier.register` and
                ``item`` is the supplied item from
                :meth:`~htf.communication.PluggableNotifier.input`.
            callback_exception_handler: a callback that is
                called to handle a raised exception in a callback.
                If the ``callback_exception_handler`` returns ``True`` the
                exception is not raised afterwards. Else the exception handler
                raises the catched exception and stopps the thread.
                The supplied handler musst accept ``name`` and ``item``
                arguments. ``name`` is the callback name supplied by
                :meth:`~htf.communication.PluggableNotifier.register` and
                ``item`` is the supplied item from
                :meth:`~htf.communication.PluggableNotifier.input`.
        """
        super(PluggableNotifier, self).__init__()

        self._verbosity = verbosity
        self._raise_exceptions = raise_exceptions
        self._filter_exception_handler = filter_exception_handler
        self._callback_exception_handler = callback_exception_handler

        self._notifications: Dict[str, Tuple[Callback, Callback]] = {}
        self._lock = threading.Lock()
        self._queue: queue.Queue = queue.Queue()

        self.daemon = True
        self._run = True
        self.start()

    def input(self, item: Any) -> None:
        """
        Input an ``item`` to be used with all registered subjects.

        Args:
            item: an item to be processed.

        Raises:
            RuntimeError: if :func:`~htf.communication.PluggableNotifier.stop`
            was called before
        """
        with self._lock:
            if not self._run:
                raise RuntimeError("You cannot input anything in a stopped " "PluggableNotifier")
        self._queue.put(item)

    def register(self, name: str, filter_expression: Callback, callback: Callback) -> None:
        """
        Register a tuple of `name`, `filter_expression` and a `callback`.

        Args:
            name: a name for the tuple (for debugging,
                see ``verbosity``).
            filter_expression: a callable filter expression
                (see :mod:`htf.filters` for more information) or
                a method that is called.
                ``filter_expression`` is called with the ``item`` from
                :func:`input` as a parameter.
                If ``filter_expression`` returns `True` the callback is called
                with the same `item`.
            callback: a callable object that is called in case
                ``filter_expression(item)`` returns ``True``.
        """
        with self._lock:
            self._notifications[name] = (filter_expression, callback)

    def unregister(self, name: str) -> None:
        """
        Unregister a callback with a given ``name``.

        Args:
            name: the name of the callback to be unregistered.

        Raises:
            KeyError: if ``name`` was not found in ``notifications``.
        """
        with self._lock:
            del self._notifications[name]

    def get_filter_names(self) -> List[str]:
        """
        Get the filter names.

        Returns:
            list of str: a sorted list containing the filter names.
        """
        with self._lock:
            names = list(sorted(self._notifications.keys()))

        return names

    def _handle_filter_exception(self, name: str, item: Any) -> None:
        if self._raise_exceptions is False:
            return
        if callable(self._filter_exception_handler):
            if self._filter_exception_handler(name, item) is True:
                return
        raise

    def _handle_callback_exception(self, name: str, item: Any) -> None:
        if self._raise_exceptions is False:
            return
        if callable(self._callback_exception_handler):
            if self._callback_exception_handler(name, item) is True:
                return
        raise

    def _running(self) -> bool:
        with self._lock:
            return self._run

    def run(self) -> None:
        while self._running():
            try:
                item = self._queue.get(timeout=0.5)

                # apply all filters on item catching exceptions
                with self._lock:
                    notifications = copy.deepcopy(self._notifications)

                for name in notifications.keys():
                    result = False
                    try:
                        result = notifications[name][0](name=name, item=item)
                    except BaseException:
                        self._handle_filter_exception(name, item)

                    if result is True:
                        try:
                            notifications[name][1](name=name, item=item)
                        except BaseException:
                            self._handle_callback_exception(name, item)

            except queue.Empty:
                pass

    def join(self, timeout: Optional[float] = 1.0) -> None:
        """
        Wait until the thread ends.

        Args:
            timeout: the timeout in seconds. Default: 1.0 seconds.
        """
        self.stop()
        super(PluggableNotifier, self).join(timeout=timeout)

    def stop(self) -> None:
        """
        Stop the thread.
        """
        with self._lock:
            self._run = False
