#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, Tuple
from htf.communication.core import SSHClientBase
import paramiko  # type: ignore


class SSHClient(SSHClientBase):
    def __init__(self) -> None:
        """
        ``SSHClient`` executes commands remotely using ssh.
        """
        SSHClientBase.__init__(self)

    def connect(self, hostname: str, username: Optional[str] = None, password: Optional[str] = None) -> None:
        """
        Establish an ssh connection.

        Args:
            hostname: the hostname or ip address
            username=None: The username to be used. If set to ``None`` the current username is used instead
             of using the ssh keys in your system.
            password=None: The password for ``username``. If set to ``None`` the password for ``username``
             or the current username is used.
        """

        self._client.connect(hostname, username=username, password=password)
        self._connected = True

    def _check_connection(self) -> None:
        if self._connected is False:
            raise Exception("Not connected. Call connect() first!")

    def close(self) -> None:
        """
        Close current ssh connection.
        """
        self._client.close()
        self._connected = False

    def execute(self, *command: str) -> Tuple[int, bytes, bytes]:
        """
        Execute a command remotely via ssh.

        Args:
            *command: the command to be executed.

        Returns:
            tuple: a tuple containing exitcode, stdout and stderr of the executed command.
        """
        self._check_connection()

        if len(command) > 1:
            cmd = " ".join(command)
        else:
            cmd = command[0]
        # print("executing command:", cmd)

        stdin, stdout, stderr = self._client.exec_command(cmd)

        return stdout.channel.recv_exit_status(), stdout.read(), stderr.read()


class SFTPClient(SSHClient):
    def __init__(self) -> None:
        """
        ``SFTPClient`` lets you access files via sftp protocol.
        """
        super(SFTPClient, self).__init__()
        self._sftpclient: Optional[paramiko.SFTPClient] = None  # type: ignore

    def connect(self, hostname: str, username: Optional[str] = None, password: Optional[str] = None) -> None:
        """
        Establish an sftp connection.

        Args:
            hostname: the hostname or ip address
            username=None: The username to be used. If set to ``None`` the current username is used instead
             of using the ssh keys in your system.
            password=None: The password for ``username``. If set to ``None`` the password for ``username``
             or the current username is used.
        """

        self._client.connect(hostname, username=username, password=password)
        self._sftpclient = self._client.open_sftp()

    def _check_connection(self) -> None:
        if self._sftpclient is None:
            raise Exception("Not connected. Call connect() first!")

    def close(self) -> None:
        """
        Close current sftp connection.
        """
        self._check_connection()
        self._client.close()
        self._sftpclient = None

    def get(self, remotepath: str, localpath: str) -> None:
        """
        Copies file identified by ``remotepath`` to ``localpath`` using sftp protocol.

        Args:
            remotepath: the remote file to be copied
            localpath: the local path where the field is copied to
        """
        self._check_connection()

        self._sftpclient.get(remotepath, localpath)  # type: ignore

    def put(self, localpath: str, remotepath: str) -> paramiko.SFTPAttributes:  # type: ignore
        """
        Copies file identified by ``localpath`` to ``remotepath`` using sftp protocol.

        Args:
            localpath: the local path to be copied
            remotepath: the remote file where the field is copied to

        Returns:
            SFTPAttributes: the result of os.stat

        """
        self._check_connection()

        return self._sftpclient.put(localpath, remotepath)  # type: ignore

    def remove(self, remotepath: str) -> None:
        """
        Removes a file identified by ``remotepath``.

        Args:
            remotepath: the path for the file to be removed

        Raises:
            IOError: if the path is not a file
        """
        self._check_connection()

        self._sftpclient.remove(remotepath)  # type: ignore

    unlink = remove

    def rename(self, oldpath: str, newpath: str) -> None:
        """
        Remotely renames a file or a folder.

        Args:
            oldpath: the file or folder to be renamed
            newpath: the new file- or foldername

        Raises:
            IOError: if something goes wrong
        """
        self._check_connection()

        self._sftpclient.rename(oldpath, newpath)  # type: ignore
